﻿using GameUtils.Graphics.Data;

namespace GameUtils.Graphics
{
    public abstract class Resource<T> : IFactoryDependentResource where T : ResourceData
    {
        readonly ResourceHandle handle;

        /// <summary>
        /// The factory-specific data of this resource.
        /// </summary>
        /// <remarks>Normally this property is for internal use only.</remarks>
        internal T Data { get; private set; }

        /// <summary>
        /// Indicates whether this resource is created.
        /// </summary>
        public bool IsCreated { get; private set; }

        /// <summary>
        /// A unique identification for this resource.
        /// </summary>
        public object Tag { get; set; }

        protected Resource()
        {
            IsCreated = false;
            handle = GameEngine.RegisterResource(this);
        }

        protected abstract T CreateData(Factory factory);

        protected abstract void ApplyChanges();

        void IFactoryDependentResource.FactoryChanged(Factory factory)
        {
            this.DisposeInner();

            if (factory != null)
            {
                Data = this.CreateData(factory);
                IsCreated = true;
                this.DataCreated(Data);
            }
        }

        void IResource.ApplyChanges()
        {
            this.ApplyChanges();
        }

        protected virtual void DataCreated(T data)
        { }

        private void DisposeInner()
        {
            IsCreated = false;

            if (Data != null)
                Data.Dispose();
        }

        public void Dispose()
        {
            handle.Dispose();
            this.DisposeInner();
        }

        ~Resource()
        {
            DisposeInner();
        }
    }
}
